import subprocess
import inquirer
import logging
import json
import os
import sys
import jwt

from .. import exceptions, entities, repositories, utilities, assets, miscellaneous

logger = logging.getLogger(name='dtlpy')


class CommandExecutor:

    def __init__(self, dl, parser):
        self.dl = dl
        self.parser = parser
        self.utils = Utils(dl)

    def run(self, args):
        ########################
        # Run Command if Exist #
        ########################
        if args.operation is None:
            logger.info('See "dlp --help" for options')
            return

        operation = args.operation.lower().replace('-', '_')
        if hasattr(self, operation):
            getattr(self, operation)(args)
        ###############
        # Catch typos #
        ###############
        elif args.operation in ["project", 'dataset', 'item', 'service', 'package', 'video', 'deploy', 'generate']:
            self.typos(args=args)
        #######################
        # Catch other options #
        #######################
        elif args.operation:
            print('dlp: "%s" is not an dlp command' % args.operation)
            print('See "dlp --help" for options')
        else:
            print('See "dlp --help" for options')

    def help(self, args):
        self.parser.print_help()

    def logout(self, args):
        self.dl.logout()
        logger.info('logout successful')

    # noinspection PyUnusedLocal
    def login(self, args):
        self.dl.login()
        self.dl.info(with_token=False)

    def login_token(self, args):
        self.dl.login_token(args.token)
        self.dl.info(with_token=False)

    def login_api_key(self, args):
        self.dl.login_api_key(api_key=args.api_key)

    def login_secret(self, args):
        self.login_m2m(args=args)

    def login_m2m(self, args):
        self.dl.login_m2m(email=args.email,
                          password=args.password,
                          client_id=args.client_id,
                          client_secret=args.client_secret)
        self.dl.info(with_token=False)

    def upgrade(self, args):
        url = 'dtlpy'
        if args.url is None:
            try:
                # oxsec-disable jwt-signature-disabled - Client-side SDK: signature verification disabled intentionally to check admin role; server validates on API calls
                payload = jwt.decode(
                    self.dl.client_api.token,
                    options={
                        "verify_signature": False,
                        "verify_exp": False,
                        "verify_aud": False,
                        "verify_iss": False,
                    }
                )
                if 'admin' in payload['https://dataloop.ai/authorization']['roles']:
                    url = "https://storage.googleapis.com/dtlpy/dev/dtlpy-latest-py3-none-any.whl"
            except Exception:
                pass
        else:
            url = args.url

        logger.info("Update DTLPy from {}".format(url))
        logger.info("Installing using pip...")
        cmd = "pip install {} --upgrade".format(url)
        subprocess.Popen(cmd, shell=True)
        sys.exit(0)

    # noinspection PyUnusedLocal
    def init(self, args):
        self.dl.init()

    # noinspection PyUnusedLocal
    def checkout_state(self, args):
        state = self.dl.checkout_state()
        logger.info('Checked-out:')
        for key, val in state.items():
            try:
                msg = '{entity} name: {name}\t{entity} id: {id}'.format(entity=key, name=val['name'], id=val['id'])
            except KeyError:
                msg = '{entity} Not Found'.format(entity=key)
            logger.info(msg)

    # noinspection PyUnusedLocal
    def version(self, args):
        logger.info("Dataloop SDK Version: {}".format(self.dl.__version__))

    def development(self, args):
        if args.development == "local":
            # start the local development
            if args.local == 'start':
                development = self.utils.dl.client_api.state_io.get('development')
                # create default values
                if development is None:
                    development = dict()
                if 'port' not in development:
                    development['port'] = 5802
                if 'docker_image' not in development:
                    development['docker_image'] = 'dataloopai/dtlpy-agent:1.57.3.gpu.cuda11.5.py3.8.opencv'

                # get values from input
                if args.docker_image is not None:
                    development['docker_image'] = args.docker_image
                if args.port is not None:
                    development['port'] = int(args.port)

                # set values to local state
                self.utils.dl.client_api.state_io.put('development', development)
                utilities.local_development.start_session()
            elif args.local == 'pause':
                utilities.local_development.pause_session()
            elif args.local == 'stop':
                utilities.local_development.stop_session()
            else:
                print('Must select one of "start", "pause", "stop". Type "dlp development start --help" for options')
        elif args.development == "remote":
            logger.warning('FUTURE! This is not supported yet..')
        else:
            print('Type "dlp development --help" for options')

    # noinspection PyUnusedLocal

    def api(self, args):
        if args.api == "info":
            information = self.dl.info()
            logger.info('-- Dataloop info --')
            _ = [logger.info('{}: {}'.format(key, val)) for key, val in information.items()]

        if args.api == "setenv":
            self.dl.setenv(args.env)
            logger.info("Platform environment: {}".format(self.dl.environment()))

    def projects(self, args):
        if args.projects == "ls":
            self.dl.projects.list().print()
        elif args.projects == "web":
            if args.project_name is None:
                args.project_name = self.dl.projects.get().name
            self.dl.projects.open_in_web(project_name=args.project_name)
        elif args.projects == "create":
            project = self.dl.projects.create(args.project_name)
            project.print()
            project.checkout()
        elif args.projects == "checkout":
            self.dl.projects.checkout(project_name=args.project_name)
        else:
            print('Type "dlp projects --help" for options')

    def datasets(self, args):
        if args.datasets == "ls":
            self.utils.get_datasets_repo(args=args).list().print()

        elif args.datasets == "checkout":
            self.dl.datasets.checkout(dataset_name=args.dataset_name)

        elif args.datasets == "web":
            if args.dataset_name is None:
                args.dataset_name = self.dl.datasets.get().name
            self.utils.get_datasets_repo(args=args).open_in_web(dataset_name=args.dataset_name)

        elif args.datasets == "create":
            dataset = self.utils.get_datasets_repo(args=args).create(dataset_name=args.dataset_name)
            dataset.print()
            if args.checkout:
                dataset.checkout()
        else:
            print('Type "dlp datasets --help" for options')

    def items(self, args):
        if self.dl.token_expired():
            logger.error("token expired, please login.")
            return

        if args.items == "ls":
            project = self.dl.projects.get(project_name=args.project_name)
            dataset = project.datasets.get(dataset_name=args.dataset_name)
            if isinstance(args.page, str):
                try:
                    args.page = int(args.page)
                except ValueError:
                    raise ValueError("Input --page must be integer")
            filters = self.dl.Filters()

            # add filters
            if args.remote_path is not None:
                if isinstance(args.remote_path, list):
                    filters.add(field="filename", values=args.remote_path, operator=entities.FiltersOperations.IN)
                else:
                    filters.add(field="filename", values=args.remote_path)
            if args.type is not None:
                if isinstance(args.type, list):
                    filters.add(field='metadata.system.mimetype', values=args.type,
                                operator=entities.FiltersOperations.IN)
                else:
                    filters.add(field="metadata.system.mimetype", values=args.type)

            pages = dataset.items.list(filters=filters, page_offset=args.page)
            pages.print()
            print("Displaying page %d/%d" % (args.page, pages.total_pages_count - 1))

        elif args.items == "web":
            project = self.dl.projects.get(project_name=args.project_name)
            dataset = project.datasets.get(dataset_name=args.dataset_name)
            dataset.items.open_in_web(filepath=args.remote_path)

        elif args.items == "upload":
            logger.info("Uploading directory...")
            if isinstance(args.file_types, str):
                args.file_types = [t.strip() for t in args.file_types.split(",")]
            project = self.dl.projects.get(project_name=args.project_name)
            dataset = project.datasets.get(dataset_name=args.dataset_name)

            # Validate local_path and local_annotations_path to prevent path traversal
            miscellaneous.PathUtils.validate_paths(
                [args.local_path, args.local_annotations_path],
                base_path=os.getcwd(),
                must_exist=True
            )

            dataset.items.upload(local_path=args.local_path,
                                 remote_path=args.remote_path,
                                 file_types=args.file_types,
                                 overwrite=args.overwrite,
                                 local_annotations_path=args.local_annotations_path)

        elif args.items == "download":
            logger.info("Downloading dataset...")
            project = self.dl.projects.get(project_name=args.project_name)
            dataset = project.datasets.get(dataset_name=args.dataset_name)

            annotation_options = None
            if args.annotation_options is not None:
                annotation_options = [t.strip() for t in args.annotation_options.split(",")]

            annotation_filters = None
            if args.annotation_filter_type is not None or args.annotation_filter_label is not None:
                annotation_filters = entities.Filters(resource=entities.FiltersResource.ANNOTATION)
                if args.annotation_filter_type is not None:
                    annotation_filter_type = [t.strip() for t in args.annotation_filter_type.split(",")]
                    annotation_filters.add(field='type',
                                           values=annotation_filter_type,
                                           operator=entities.FiltersOperations.IN)
                if args.annotation_filter_label is not None:
                    annotation_filter_label = [t.strip() for t in args.annotation_filter_label.split(",")]
                    annotation_filters.add(field='label',
                                           values=annotation_filter_label,
                                           operator=entities.FiltersOperations.IN)
            # create remote path filters
            filters = self.dl.Filters()
            if args.remote_path is not None:
                remote_path = [t.strip() for t in args.remote_path.split(",")]
                if len(remote_path) == 1:
                    remote_path = remote_path[0]
                    filters.add(field="filename", values=remote_path)
                elif len(remote_path) > 1:
                    for item in remote_path:
                        if '*' in item:
                            filters.add(field="dir", values=item, method='or')
                            remote_path.pop(remote_path.index(item))
                    filters.add(field="dir", values=remote_path, operator=entities.FiltersOperations.IN, method='or')

            # Validate local_path to prevent path traversal
            miscellaneous.PathUtils.validate_directory_path(
                args.local_path,
                base_path=os.getcwd(),
                must_exist=False
            )

            if not args.without_binaries:
                dataset.items.download(filters=filters,
                                       local_path=args.local_path,
                                       annotation_options=annotation_options,
                                       annotation_filters=annotation_filters,
                                       overwrite=args.overwrite,
                                       with_text=args.with_text,
                                       thickness=int(args.thickness),
                                       to_items_folder=not args.not_items_folder)
            else:
                dataset.download_annotations(filters=filters,
                                             local_path=args.local_path,
                                             annotation_options=annotation_options,
                                             annotation_filters=annotation_filters,
                                             overwrite=args.overwrite,
                                             with_text=args.with_text,
                                             thickness=int(args.thickness))

        else:
            print('Type "dlp items --help" for options')

    def videos(self, args):
        if self.dl.token_expired():
            print("[ERROR] token expired, please login.")
            return

        if args.videos == "play":
            from dtlpy.utilities.videos.video_player import VideoPlayer

            VideoPlayer(
                item_filepath=args.item_path,
                dataset_name=args.dataset_name,
                project_name=args.project_name,
            )

        elif args.videos == "upload":
            if (
                    (args.split_chunks is not None)
                    or (args.split_seconds is not None)
                    or (args.split_times is not None)
            ):
                # upload with split
                if isinstance(args.split_chunks, str):
                    args.split_chunks = int(args.split_chunks)
                if isinstance(args.split_seconds, str):
                    args.split_seconds = int(args.split_seconds)
                if isinstance(args.split_times, str):
                    args.split_times = [int(sec) for sec in args.split_times.split(",")]
                # Validate filepath to prevent path traversal
                miscellaneous.PathUtils.validate_file_path(args.filename)
                
                self.dl.utilities.videos.Videos.split_and_upload(
                    project_name=args.project_name,
                    dataset_name=args.dataset_name,
                    filepath=args.filename,
                    remote_path=args.remote_path,
                    split_chunks=args.split_chunks,
                    split_seconds=args.split_seconds,
                    split_pairs=args.split_times,
                )
        else:
            print('Type "dlp videos --help" for options')

    def services(self, args):
        if self.dl.token_expired():
            print("[ERROR] token expired, please login.")
            return

        elif args.services == "delete":
            service = self.utils.get_services_repo(args=args).get(service_name=args.service_name)
            service.delete()
            logger.info('Service: "{}" deleted successfully'.format(service.name))

        elif args.services == "ls":
            self.utils.get_services_repo(args=args).list().print()
        elif args.services == "log":
            project = self.dl.projects.get(project_name=args.project_name)
            service = project.services.get(service_name=args.service_name)

            logs = service.log(start=args.start)
            try:
                for log in logs:
                    if isinstance(log, list):
                        for log_record in log:
                            print(log_record)
                    else:
                        print(log)
            except KeyboardInterrupt:
                pass

        elif args.services == "execute":
            try:
                if args.service_name:
                    service = self.utils.get_services_repo(args=args).get(service_name=args.service_name)
                else:
                    service = self.dl.services.get()
            except Exception:
                logger.info('Please checkout a service')
                return

            if args.annotation_id is not None:
                resource = 'Annotation'
            elif args.item_id is not None:
                resource = 'Item'
            elif args.dataset_id is not None:
                resource = 'Dataset'
            else:
                resource = None

            try:
                execution_input = json.loads(args.inputs)
            except Exception:
                logger.info('Input should be json serializable')
                return
            if len(execution_input) == 0:
                execution_input = None

            logger.info('executing')
            service.execute(sync=not args.asynchronous,
                            execution_input=execution_input,
                            function_name=args.function_name,
                            resource=resource,
                            item_id=args.item_id,
                            dataset_id=args.dataset_id,
                            annotation_id=args.annotation_id)
            logger.info("Successfully executed function: {}".format(args.function_name))

        else:
            logger.info('Type "dlp packages --help" for options')

    def deploy(self, args):
        project = self.dl.projects.get(project_name=args.project_name)
        json_filepath = args.json_file
        # Validate file path to prevent path traversal
        miscellaneous.PathUtils.validate_file_path(json_filepath)
        deployed_services, package = self.dl.packages.deploy_from_file(project=project, json_filepath=json_filepath)
        logger.info("Successfully deployed {} from file: {}\nServices: {}".format(len(deployed_services),
                                                                                  json_filepath,
                                                                                  [s.name for s in deployed_services]))

    def generate(self, args):
        package_type = args.package_type if args.package_type else self.dl.PackageCatalog.DEFAULT_PACKAGE_TYPE
        self.dl.packages.generate(name=args.package_name, src_path=os.getcwd(), package_type=package_type)
        self.utils.dl.client_api.state_io.put('package', {'name': args.package_name})
        logger.info('Successfully generated package files')

    def triggers(self, args):

        if args.triggers == "create":
            args.actions = [t.strip() for t in args.actions.split(",")]
            try:
                service = self.utils.get_services_repo(args).get(service_name=args.service_name)
            except exceptions.NotFound:
                logger.critical('Service not found. Please check-out a service or provide valid service name')
                return

            trigger = service.triggers.create(name=args.name,
                                              filters=json.loads('{}'.format(args.filters.replace("'", '"'))),
                                              function_name=args.function_name,
                                              resource=args.resource,
                                              actions=args.actions)
            logger.info('Trigger created successfully: {}'.format(trigger.name))

        elif args.triggers == "delete":
            triggers = self.utils.get_triggers_repo(args=args)
            triggers.get(trigger_name=args.trigger_name).delete()
            logger.info('Trigger deleted successfully: {}'.format(args.trigger_name))
        elif args.triggers == "ls":
            triggers = self.utils.get_triggers_repo(args=args)
            triggers.list().print()
        else:
            logger.info('Type "dlp packages --help" for options')

    def packages(self, args):
        if self.dl.token_expired():
            logger.error("token expired, please login.")
            return

        elif args.packages == "delete":
            package = self.utils.get_packages_repo(args=args).get(package_name=args.package_name)
            package.delete()
            logger.info('Successfully deleted package {}'.format(package.name))

        elif args.packages == "ls":
            self.utils.get_packages_repo(args=args).list().print()

        elif args.packages == "checkout":
            self.dl.packages.checkout(package_name=args.package_name)

        elif args.packages == "push":
            packages = self.utils.get_packages_repo(args=args)

            # Validate src_path to prevent path traversal
            miscellaneous.PathUtils.validate_directory_path(
                args.src_path,
                base_path=os.getcwd(),
                must_exist=True
            )

            package = packages.push(src_path=args.src_path,
                                    package_name=args.package_name,
                                    checkout=args.checkout)

            logger.info("Successfully pushed package to platform\n"
                        "Package id:{}\nPackage version:{}".format(package.id,
                                                                   package.version))
        elif args.packages == "deploy":
            packages = self.utils.get_packages_repo(args=args)

            package = packages.deploy(package_name=args.package_name,
                                      checkout=args.checkout,
                                      module_name=args.module_name)

            logger.info("Successfully pushed package to platform\n"
                        "Package id:{}\nPackage version:{}".format(package.id,
                                                                   package.version))

        elif args.packages == "test":
            go_back = False
            if 'src' in os.listdir(os.getcwd()):
                go_back = True
                os.chdir('./src')
            try:
                self.dl.packages.test_local_package(concurrency=int(args.concurrency),
                                                    function_name=args.function_name)
            except Exception:
                logger.exception('failed during test')
            finally:
                if go_back:
                    os.chdir('..')

        else:
            logger.info('Type "dlp packages --help" for options')

    def app(self, args):
        if args.app == 'pack':
            path = self.dl.dpks.pack()
            logger.info(f'Packed to {path}')
        elif args.app == 'init':
            app_filename = assets.paths.APP_JSON_FILENAME
            if os.path.isfile(app_filename):
                questions = [
                    inquirer.Confirm(name='overwrite',
                                     message=f"Dataloop app already initialized. Re-initialize?",
                                     default=False)]
                answers = inquirer.prompt(questions)
                if answers.get('overwrite') is False:
                    return
            name = args.name
            description = args.description
            attributes = args.attributes
            icon = args.icon
            scope = args.scope
            as_array = [name, description, attributes, icon, scope]
            if as_array.count(None) == len(as_array):  # No one value is initialized
                dir_name = os.path.basename(os.getcwd())
                questions = [
                    inquirer.Text(name='name',
                                  message=f"Enter the name of the app (or press enter for '{dir_name}'):",
                                  default=dir_name),
                    inquirer.Text(name='description',
                                  message="Enter the description (or enter for empty): "),
                    inquirer.Text(name='attributes',
                                  message="Enter the attributes (an object, or enter for empty): ",
                                  default=None),
                    inquirer.Text(name='icon',
                                  message="Enter the path to the icon (or enter for empty): "),
                    inquirer.Text(name='scope',
                                  message="Enter the scope (or enter for 'organization'): ",
                                  default='project'),
                ]
                answers = inquirer.prompt(questions)
                name = answers.get('name')
                description = answers.get('description')
                attributes = answers.get('attributes')
                icon = answers.get('icon')
                scope = answers.get('scope')

            if attributes is not None:
                attributes = json.loads(attributes)

            self.dl.dpks.init(name=name,
                              description=description,
                              attributes=attributes,
                              icon=icon,
                              scope=scope)
        elif args.app == 'add':
            if args.panel is True:
                default_panel_name = "myPanel"
                choices = list(entities.dpk.SlotType)

                questions = [
                    inquirer.Text(name='name',
                                  message=f"Enter PANEL NAME (or press enter for '{default_panel_name}'): ",
                                  default='default_panel_name'),
                    inquirer.List(name='support_slot_type',
                                  message="Enter SUPPORTED SLOT TYPE:",
                                  choices=choices),
                ]

                answers = inquirer.prompt(questions)
                #####
                # create a dir for that panel
                # Validate panel name to prevent path traversal
                panel_name = answers.get('name')
                # Validate panel name to prevent path traversal
                miscellaneous.PathUtils.validate_directory_name(panel_name)
                # Create directory in current working directory
                panel_dir = os.path.join(os.getcwd(), panel_name)
                os.makedirs(panel_dir, exist_ok=True)
                # dump to dataloop.json
                app_filename = assets.paths.APP_JSON_FILENAME
                if not os.path.isfile(app_filename):
                    logger.error(f"Can't find app config file ({app_filename}), please run `dlp app init` first")
                else:
                    with open(app_filename, 'r') as f:
                        dpk = entities.Dpk.from_json(json.load(f))
                    dpk.components.panels.append(entities.Panel(name=answers.get('name'),
                                                                supported_slots=[answers.get('support_slot_type')]))
                    with open(app_filename, 'w') as f:
                        json.dump(dpk.to_json(), f, indent=4)

        elif args.app == 'publish':
            dpk = self.utils.get_dpks_repo(args).publish()
            if dpk:
                logger.info(f'Published the application: id={dpk.id}')
            else:
                logger.info("Couldn't publish the application")
        elif args.app == 'update':
            # TODO: I think it changed, not implemented
            logger.info('App updated successfully')
        elif args.app == 'resume':
            succeed = self.utils.get_apps_repo(args).resume(app_id=args.app_id)
            if succeed is True:
                logger.info('Resumed application successfully')
            else:
                logger.info('Application resume failed')
        elif args.app == 'pause':
            succeed = self.utils.get_apps_repo(args).pause(app_id=args.app_id)
            if succeed is True:
                logger.info('Paused application successfully')
            else:
                logger.info('Application pause failed')
        elif args.app == 'install':
            app = self.utils.get_apps_repo(args).install(
                dpk=self.utils.get_dpks_repo(args).get(dpk_id=args.dpk_id),
                organization_id=args.org_id
            )
            if app is not None:
                logger.info('App installed successfully')
            else:
                logger.info("Couldn't install the app")
        elif args.app == 'pull':
            succeed = self.dl.dpks.pull(dpk_name=args.dpk_name)
            if succeed is True:
                logger.info("Pulled successfully")
            else:
                logger.info("Couldn't pull")
        elif args.app == 'list':
            self.dl.apps.list().print()

    # noinspection PyUnusedLocal
    @staticmethod
    def pwd(args):
        print(os.getcwd())

    @staticmethod
    def cd(args):
        directory = args.dir
        if directory == '..':
            directory = os.path.split(os.getcwd())[0]
        # Validate path to prevent path traversal
        miscellaneous.PathUtils.validate_directory_path(
            directory,
            base_path=os.getcwd(),
            must_exist=True
        )
        os.chdir(directory)
        print(os.getcwd())

    @staticmethod
    def mkdir(args):
        # Validate directory name to prevent path traversal
        miscellaneous.PathUtils.validate_directory_name(args.name)
        # Create directory in current working directory
        dir_path = os.path.join(os.getcwd(), args.name)
        os.mkdir(dir_path)

    # noinspection PyUnusedLocal
    @staticmethod
    def ls(args):
        print(os.getcwd())
        dirs = os.listdir(os.getcwd())
        import pprint

        pp = pprint.PrettyPrinter(indent=3)
        pp.pprint(dirs)

    # noinspection PyUnusedLocal
    @staticmethod
    def clear(args):
        if os.name == "nt":
            os.system('cls')
        else:
            os.system('clear')

    @staticmethod
    def typos(args):
        print('dlp: "{op}" is not an dlp command. Did you mean "{op}s"?'.format(op=args.operation))


class Utils:
    def __init__(self, dl):
        self.dl = dl

    def get_packages_repo(self, args):
        if args.project_name is not None:
            project = self.dl.projects.get(project_name=args.project_name)
            packages = project.packages
        else:
            try:
                packages = self.dl.projects.get().packages
            except Exception:
                packages = self.dl.packages

        assert isinstance(packages, repositories.Packages)
        return packages

    def get_datasets_repo(self, args):
        if args.project_name is not None:
            project = self.dl.projects.get(project_name=args.project_name)
            datasets = project.datasets
        else:
            try:
                datasets = self.dl.projects.get().datasets
            except Exception:
                datasets = self.dl.datasets

        assert isinstance(datasets, repositories.Datasets)
        return datasets

    def get_services_repo(self, args):
        if args.project_name is not None:
            project = self.dl.projects.get(project_name=args.project_name)
            packages = project.packages
            services = project.services
        else:
            try:
                packages = self.dl.projects.get().packages
                services = self.dl.projects.get().services
            except Exception:
                packages = self.dl.packages
                services = self.dl.services

        if args.package_name is not None:
            package = packages.get(package_name=args.package_name)
            services = package.services

        assert isinstance(services, repositories.Services)
        return services

    def get_apps_repo(self, args) -> repositories.Apps:
        if args.project_name is not None:
            project = self.dl.projects.get(project_name=args.project_name)
            apps = project.apps
        else:
            try:
                apps = self.dl.projects.get().apps
            except Exception:
                apps = self.dl.apps
        assert isinstance(apps, repositories.Apps)
        return apps

    def get_dpks_repo(self, args) -> repositories.Dpks:
        if args.project_name is None and args.project_id is None:
            try:
                dpks = self.dl.projects.get().dpks
            except Exception:
                dpks = self.dl.dpks
        else:
            project = self.dl.projects.get(project_name=args.project_name, project_id=args.project_id)
            dpks = project.dpks
        if dpks.project is None:
            raise ValueError("Must input one of `project-name` or `project-id`") from None
        assert isinstance(dpks, repositories.Dpks)
        return dpks

    def get_triggers_repo(self, args):
        if args.project_name is not None:
            project = self.dl.projects.get(project_name=args.project_name)
            packages = project.packages
            services = project.services
            triggers = project.triggers
        else:
            try:
                packages = self.dl.projects.get().packages
                services = self.dl.projects.get().services
                triggers = self.dl.projects.get().triggers
            except Exception:
                packages = self.dl.packages
                services = self.dl.services
                triggers = self.dl.triggers

        if args.package_name is not None:
            package = packages.package_name(name=args.package_name)
            services = package.services

        if args.service_name is not None:
            service = services.get(service_name=args.service_name)
            triggers = service.triggers

        assert isinstance(triggers, repositories.Triggers)
        return triggers



import copy
import io
import json
import logging
import multiprocessing
import os
import shutil
import sys
import tempfile
import traceback
from pathlib import Path
from urllib.parse import unquote, urlparse

import numpy as np
import requests
import tqdm
from PIL import Image
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from .. import entities, repositories, miscellaneous, PlatformException, exceptions
from ..services import Reporter

logger = logging.getLogger(name='dtlpy')

NUM_TRIES = 3  # try to download 3 time before fail on item
DOWNLOAD_MAX_ITEMS_PER_SUBSET = 1000

class Downloader:
    def __init__(self, items_repository):
        self.items_repository = items_repository

    def _process_download_results(self, reporter, raise_on_error=False):
        """
        Process download results and generate summary report.

        :param reporter: Reporter instance containing download results
        :param raise_on_error: If True, raise exception on download errors
        :return: Output from reporter
        """
        # reporting
        n_download = reporter.status_count(status='download')
        n_exist = reporter.status_count(status='exist')
        n_error = reporter.status_count(status='error')
        logger.info(f"Number of files downloaded:{n_download}")
        logger.info(f"Number of files exists: {n_exist}")
        logger.info(f"Total number of files: {n_download + n_exist}")

        # log error
        if n_error > 0:
            log_filepath = reporter.generate_log_files()
            # Get up to 5 error examples for the exception message
            error_text = ""
            error_counter = 0
            if reporter._errors:
                for _id, error in reporter._errors.items():
                    error_counter += 1
                    error_text += f"Item ID: {_id}, Error: {error} | "
                    if error_counter >= 5:
                        break
            error_message = f"Errors in {n_error} files. Errors: {error_text}"
            if log_filepath is not None:
                error_message += f", see {log_filepath} for full log"
            if raise_on_error is True:
                raise PlatformException(
                    error="400", message=error_message
                )
            else:
                logger.warning(error_message)
        
        if int(n_download) <= 1 and int(n_exist) <= 1:
            try:
                return next(reporter.output)
            except StopIteration:
                return None
        return reporter.output

    def _process_item_json(self, local_path, item_json, reporter, pbar, overwrite=False):
        """
        Process a single item JSON for download, saving both the item file and metadata.

        :param local_path: Local path to save files
        :param item_json: Item JSON metadata
        :param reporter: Reporter instance for tracking progress
        :param pbar: Progress bar instance
        :param overwrite: Whether to overwrite existing files
        :return: Error message, traceback, and downloaded filepath
        """
        err = None
        trace = None
        downloaded_filepath = None
        item_id = item_json['id']
        filename = item_json['filename'].lstrip('/')
        
        for i_try in range(NUM_TRIES):
            try:
                # Download the image
                image_path = Path(local_path) / 'items' / filename
                # Ensure the directory for the image file exists (in case filename has subdirectories)
                image_path.parent.mkdir(parents=True, exist_ok=True)
                item = entities.Item.from_json(_json = item_json, client_api=self.items_repository._client_api, is_fetched=False)
                downloaded_data = self.__thread_download(
                    item=item,
                    local_path=str(image_path.parent),
                    local_filepath=str(image_path),
                    save_locally=True,
                    to_array=False,
                    overwrite=overwrite,
                    annotation_options=[],
                    annotation_filters=None,
                )

                if downloaded_data is None:
                    err = 'Failed to download image'
                    trace = ''
                else:
                    # Save the item JSON directly
                    json_filename = Path(filename).stem + '.json'
                    json_path = Path(local_path) / 'json' / Path(filename).parent / json_filename

                    # Ensure the directory for the JSON file exists (in case filename has subdirectories)
                    json_path.parent.mkdir(parents=True, exist_ok=True)

                    # Save the original item_json directly
                    with open(json_path, 'w', encoding='utf-8') as f:
                        json.dump(item_json, f, indent=2, ensure_ascii=False)

                    downloaded_filepath = str(image_path)
                
                if downloaded_filepath is not None:
                    break

            except Exception as e:
                logger.debug(f"Download item: {filename}. Try {i_try + 1}/{NUM_TRIES}. Fail.")
                err = e
                trace = traceback.format_exc()

        pbar.update()
        if downloaded_filepath is None:
            if err is None:
                err = self.items_repository._client_api.platform_exception
            reporter.set_index(status="error", ref=item_id, success=False, error=f"{err}\n{trace}")
        else:
            reporter.set_index(ref=item_id, status="download", output=downloaded_filepath, success=True)

    def _download_recursive(
        self,
        local_path=None,
        filters: entities.Filters = None,
        annotation_filters: entities.Filters = None,
        file_types=None,
        overwrite=False,
        raise_on_error=False,
        dataset_lock=False,
        lock_timeout_sec=None,
    ):
        """
        Download items recursively from a dataset.
        
        :param local_path: Local path to save downloaded items
        :param filters: Filters entity to filter items
        :param annotation_filters: Filters entity to filter annotations
        :param file_types: List of file types to download
        :param overwrite: Whether to overwrite existing files
        :param raise_on_error: Raise error if download fails
        :param dataset_lock: Lock dataset during download
        :param lock_timeout_sec: Lock timeout in seconds
        """
        filters, annotation_filters = self._prepare_filters(filters=filters,annotation_filters=annotation_filters,file_types=file_types)
        filter_copy = copy.deepcopy(filters)
        filter_copy.page_size = 0
        num_items = self.items_repository.list(filters=filter_copy).items_count
        if num_items == 0:
            return list()
        client_api = self.items_repository._client_api
        reporter = Reporter(
            num_workers=num_items,
            resource=Reporter.ITEMS_DOWNLOAD,
            print_error_logs=client_api.verbose.print_error_logs,
            client_api=client_api,
        )

        # Create directories once using pathlib
        local_path_obj = Path(local_path)
        items_dir = local_path_obj / 'items'
        jsons_dir = local_path_obj / 'json'
        items_dir.mkdir(parents=True, exist_ok=True)
        jsons_dir.mkdir(parents=True, exist_ok=True)

        jobs = [None for _ in range(num_items)]
        # crrently keep the thread count to default.
        # client_api._thread_pools_names['item.download'] = 5 * multiprocessing.cpu_count()
        pool = client_api.thread_pools(pool_name='item.download')
        pbar = tqdm.tqdm(
            total=num_items,
            disable=client_api.verbose.disable_progress_bar_download_dataset,
            file=sys.stdout,
            desc='Download Items',
        )
        try:
            i_item = 0
            import time
            start_time = time.time()
            for json_file in self.items_repository.dataset.project.datasets._export_recursive(
                dataset=self.items_repository.dataset,
                local_path=tempfile.mkdtemp(prefix='download_recursive_jsons_'),
                max_items_per_subset=DOWNLOAD_MAX_ITEMS_PER_SUBSET,
                include_annotations=True,
                filters=filters,
                annotation_filters=annotation_filters,
                dataset_lock=dataset_lock,
                lock_timeout_sec=lock_timeout_sec,
            ):
                end_time = time.time()
                with open(json_file, 'r') as f:
                    data = json.load(f)
                for item_json in data:
                    jobs[i_item] = pool.submit(
                        self._process_item_json,
                        **{
                            "local_path": local_path,
                            "item_json": item_json,
                            "reporter": reporter,
                            "pbar": pbar,
                            "overwrite": overwrite,
                        },
                    )
                    i_item += 1
        finally:
            _ = [j.result() for j in jobs if j is not None]
            pbar.close()
        return self._process_download_results(reporter=reporter, raise_on_error=raise_on_error)

    @staticmethod
    def _prepare_filters(filters: entities.Filters = None,
                        annotation_filters: entities.Filters = None,
                        file_types=None):
        """
        Prepare and merge filters with annotation filters.

        :param filters: Filters entity or None
        :param annotation_filters: Annotation filters to merge with item filters
        :param file_types: List of file types to filter
        :return: Prepared filters entity
        """
        # filters
        if filters is None:
            filters = entities.Filters()
            filters._user_query = 'false'
        # file types
        if file_types is not None:
            filters.add(field='metadata.system.mimetype', values=file_types, operator=entities.FiltersOperations.IN)
        if annotation_filters is not None:
            if len(annotation_filters.and_filter_list) > 0 or len(annotation_filters.or_filter_list) > 0:
                for annotation_filter_and in annotation_filters.and_filter_list:
                    filters.add_join(field=annotation_filter_and.field,
                                    values=annotation_filter_and.values,
                                    operator=annotation_filter_and.operator,
                                    method=entities.FiltersMethod.AND)
                for annotation_filter_or in annotation_filters.or_filter_list:
                    filters.add_join(field=annotation_filter_or.field,
                                    values=annotation_filter_or.values,
                                    operator=annotation_filter_or.operator,
                                    method=entities.FiltersMethod.OR)
            elif annotation_filters.custom_filter is not None:
                annotation_query_dict = annotation_filters.prepare()
                items_query_dict = filters.prepare()
                items_query_dict["join"] = annotation_query_dict
                filters.reset()
                filters.custom_filter = items_query_dict

        else:
            annotation_filters = entities.Filters(resource=entities.FiltersResource.ANNOTATION)
            filters._user_query = 'false'

        return filters, annotation_filters

    def download(self,
                 # filter options
                 filters: entities.Filters = None,
                 items=None,
                 # download options
                 local_path=None,
                 file_types=None,
                 save_locally=True,
                 to_array=False,
                 overwrite=False,
                 annotation_filters: entities.Filters = None,
                 annotation_options: entities.ViewAnnotationOptions = None,
                 to_items_folder=True,
                 thickness=1,
                 with_text=False,
                 without_relative_path=None,
                 avoid_unnecessary_annotation_download=False,
                 include_annotations_in_output=True,
                 export_png_files=False,
                 filter_output_annotations=False,
                 alpha=1,
                 export_version=entities.ExportVersion.V1,
                 dataset_lock=False,
                 lock_timeout_sec=None,
                 export_summary=False,
                 raise_on_error=False
                 ):
        """
        Download dataset by filters.
        Filtering the dataset for items and save them local
        Optional - also download annotation, mask, instance and image mask of the item

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param items: download Item entity or item_id (or a list of item)
        :param local_path: local folder or filename to save to.
        :param file_types: a list of file type to download. e.g ['video/webm', 'video/mp4', 'image/jpeg', 'image/png']
        :param save_locally: bool. save to disk or return a buffer
        :param to_array: returns Ndarray when True and local_path = False
        :param overwrite: optional - default = False
        :param annotation_options: download annotations options. options: list(dl.ViewAnnotationOptions)
        :param annotation_filters: Filters entity to filter annotations for download
        :param to_items_folder: Create 'items' folder and download items to it
        :param with_text: optional - add text to annotations, default = False
        :param thickness: optional - line thickness, if -1 annotation will be filled, default =1
        :param without_relative_path: bool - download items without the relative path from platform
        :param avoid_unnecessary_annotation_download: DEPRECATED only items and annotations in filters are downloaded
        :param include_annotations_in_output: default - False , if export should contain annotations
        :param export_png_files: default - True, if semantic annotations should be exported as png files
        :param filter_output_annotations: default - False, given an export by filter - determine if to filter out annotations
        :param alpha: opacity value [0 1], default 1
        :param str export_version:  exported items will have original extension in filename, `V1` - no original extension in filenames
        :param bool dataset_lock: optional - default = False
        :param bool export_summary: optional - default = False
        :param int lock_timeout_sec: optional
        :param bool raise_on_error: raise an exception if an error occurs
        :return: Output (list)
        """

        ###################
        # Default options #
        ###################
        # annotation options
        if annotation_options is None:
            annotation_options = list()
        elif not isinstance(annotation_options, list):
            annotation_options = [annotation_options]
        for ann_option in annotation_options:
            if not isinstance(ann_option, entities.ViewAnnotationOptions):
                if ann_option not in list(entities.ViewAnnotationOptions):
                    raise PlatformException(
                        error='400',
                        message='Unknown annotation download option: {}, please choose from: {}'.format(
                            ann_option, list(entities.ViewAnnotationOptions)))
        # normalize items argument: treat empty list as “no items specified”
        if isinstance(items, list) and len(items) == 0:
            items = None
        #####################
        # items to download #
        #####################
        if items is not None:
            # convert input to a list
            if not isinstance(items, list):
                items = [items]
            # get items by id
            if isinstance(items[0], str):
                items = [self.items_repository.get(item_id=item_id) for item_id in items]
            elif isinstance(items[0], entities.Item):
                pass
            else:
                raise PlatformException(
                    error="400",
                    message='Unknown items type to download. Expecting str or Item entities. Got "{}" instead'.format(
                        type(items[0])
                    )
                )
            # create filters to download annotations
            filters = entities.Filters(field='id',
                                       values=[item.id for item in items],
                                       operator=entities.FiltersOperations.IN)
            filters._user_query = 'false'

            # convert to list of list (like pages and page)
            items_to_download = [items]
            num_items = len(items)
        else:
            # Prepare and merge filters
            filters, annotation_filters = self._prepare_filters(
                filters=filters,
                annotation_filters=annotation_filters,
                file_types=file_types
            )

            items_to_download = self.items_repository.list(filters=filters)
            num_items = items_to_download.items_count

        if num_items == 0:
            logger.warning('No items found! Nothing was downloaded')
            return list()

        ##############
        # local path #
        ##############
        is_folder = False
        if local_path is None:
            # create default local path
            local_path = self.__default_local_path()

        if os.path.isdir(local_path):
            logger.info('Local folder already exists:{}. merge/overwrite according to "overwrite option"'.format(
                local_path))
            is_folder = True
        else:
            # check if filename
            _, ext = os.path.splitext(local_path)
            if num_items > 1:
                is_folder = True
            else:
                item_to_download = items_to_download[0][0]
                file_name = item_to_download.name
                _, ext_download = os.path.splitext(file_name)
                if ext_download != ext:
                    is_folder = True
            if is_folder and save_locally:
                path_to_create = local_path
                if local_path.endswith('*'):
                    path_to_create = os.path.dirname(local_path)
                logger.info("Creating new directory for download: {}".format(path_to_create))
                os.makedirs(path_to_create, exist_ok=True)

        ####################
        # annotations json #
        ####################
        # download annotations' json files in a new thread
        # items will start downloading and if json not exists yet - will download for each file
        if num_items > 1 and annotation_options:
            # a new folder named 'json' will be created under the "local_path"
            logger.info("Downloading annotations formats: {}".format(annotation_options))
            self.download_annotations(**{
                "dataset": self.items_repository.dataset,
                "filters": filters,
                "annotation_filters": annotation_filters,
                "local_path": local_path,
                'overwrite': overwrite,
                'include_annotations_in_output': include_annotations_in_output,
                'export_png_files': export_png_files,
                'filter_output_annotations': filter_output_annotations,
                'export_version': export_version,
                'dataset_lock': dataset_lock,
                'lock_timeout_sec': lock_timeout_sec,
                'export_summary': export_summary
            })
        ###############
        # downloading #
        ###############
        # create result lists
        client_api = self.items_repository._client_api

        reporter = Reporter(num_workers=num_items,
                            resource=Reporter.ITEMS_DOWNLOAD,
                            print_error_logs=client_api.verbose.print_error_logs,
                            client_api=client_api)
        jobs = [None for _ in range(num_items)]
        # pool
        pool = client_api.thread_pools(pool_name='item.download')
        # download
        pbar = tqdm.tqdm(total=num_items, disable=client_api.verbose.disable_progress_bar_download_dataset,
                         file=sys.stdout,
                         desc='Download Items')
        try:
            i_item = 0
            for page in items_to_download:
                for item in page:
                    if item.type == "dir":
                        continue
                    if save_locally:
                        # get local file path
                        item_local_path, item_local_filepath = self.__get_local_filepath(
                            local_path=local_path,
                            without_relative_path=without_relative_path,
                            item=item,
                            to_items_folder=to_items_folder,
                            is_folder=is_folder)

                        if os.path.isfile(item_local_filepath) and not overwrite:
                            logger.debug("File Exists: {}".format(item_local_filepath))
                            reporter.set_index(ref=item.id, status='exist', output=item_local_filepath, success=True)
                            pbar.update()
                            if annotation_options and item.annotated:
                                # download annotations only
                                jobs[i_item] = pool.submit(
                                    self._download_img_annotations,
                                    **{
                                        "item": item,
                                        "img_filepath": item_local_filepath,
                                        "overwrite": overwrite,
                                        "annotation_options": annotation_options,
                                        "annotation_filters": annotation_filters,
                                        "local_path": item_local_path,
                                        "thickness": thickness,
                                        "alpha": alpha,
                                        "with_text": with_text,
                                        "export_version": export_version,
                                    },
                                )
                            i_item += 1
                            continue
                    else:
                        item_local_path = None
                        item_local_filepath = None

                    # download single item
                    jobs[i_item] = pool.submit(
                        self.__thread_download_wrapper,
                        **{
                            "i_item": i_item,
                            "item": item,
                            "item_local_path": item_local_path,
                            "item_local_filepath": item_local_filepath,
                            "save_locally": save_locally,
                            "to_array": to_array,
                            "annotation_options": annotation_options,
                            "annotation_filters": annotation_filters,
                            "reporter": reporter,
                            "pbar": pbar,
                            "overwrite": overwrite,
                            "thickness": thickness,
                            "alpha": alpha,
                            "with_text": with_text,
                            "export_version": export_version
                        },
                    )
                    i_item += 1
        except Exception:
            logger.exception('Error downloading:')
        finally:
            _ = [j.result() for j in jobs if j is not None]
            pbar.close()

        return self._process_download_results(reporter=reporter, raise_on_error=raise_on_error)

    def __thread_download_wrapper(self, i_item,
                                  # item params
                                  item, item_local_path, item_local_filepath,
                                  save_locally, to_array, overwrite,
                                  # annotations params
                                  annotation_options, annotation_filters, with_text, thickness,
                                  # threading params
                                  reporter, pbar, alpha, export_version):

        download = None
        err = None
        trace = None
        for i_try in range(NUM_TRIES):
            try:
                logger.debug("Download item: {path}. Try {i}/{n}. Starting..".format(path=item.filename,
                                                                                     i=i_try + 1,
                                                                                     n=NUM_TRIES))
                download = self.__thread_download(item=item,
                                                  save_locally=save_locally,
                                                  to_array=to_array,
                                                  local_path=item_local_path,
                                                  local_filepath=item_local_filepath,
                                                  annotation_options=annotation_options,
                                                  annotation_filters=annotation_filters,
                                                  overwrite=overwrite,
                                                  thickness=thickness,
                                                  alpha=alpha,
                                                  with_text=with_text,
                                                  export_version=export_version)
                logger.debug("Download item: {path}. Try {i}/{n}. Success. Item id: {id}".format(path=item.filename,
                                                                                                 i=i_try + 1,
                                                                                                 n=NUM_TRIES,
                                                                                                 id=item.id))
                if download is not None:
                    break
            except Exception as e:
                logger.debug("Download item: {path}. Try {i}/{n}. Fail.".format(path=item.filename,
                                                                                i=i_try + 1,
                                                                                n=NUM_TRIES))
                err = e
                trace = traceback.format_exc()
        pbar.update()
        if download is None:
            if err is None:
                err = self.items_repository._client_api.platform_exception
            reporter.set_index(status="error", ref=item.id, success=False,
                               error="{}\n{}".format(err, trace))
        else:
            reporter.set_index(ref=item.id, status="download", output=download, success=True)

    @staticmethod
    def download_annotations(dataset: entities.Dataset,
                             local_path: str,
                             filters: entities.Filters = None,
                             annotation_filters: entities.Filters = None,
                             overwrite=False,
                             include_annotations_in_output=True,
                             export_png_files=False,
                             filter_output_annotations=False,
                             export_version=entities.ExportVersion.V1,
                             dataset_lock=False,
                             lock_timeout_sec=None,
                             export_summary=False
                             ):
        """
        Download annotations json for entire dataset

        :param dataset: Dataset entity
        :param local_path:
        :param dtlpy.entities.filters.Filters filters: dl.Filters entity to filters items
        :param annotation_filters: dl.Filters entity to filters items' annotations
        :param overwrite: optional - overwrite annotations if exist, default = false
        :param include_annotations_in_output: default - True , if export should contain annotations
        :param export_png_files: default - if True, semantic annotations should be exported as png files
        :param filter_output_annotations: default - False, given an export by filter - determine if to filter out annotations
        :param str export_version:  exported items will have original extension in filename, `V1` - no original extension in filenames
        :param bool dataset_lock: optional - default = False
        :param bool export_summary: optional - default = False
        :param int lock_timeout_sec: optional
        :return:
        """
        local_path = os.path.join(local_path, "json")
        zip_filepath = None
        # only if json folder does not exist or exist and overwrite
        if not os.path.isdir(os.path.join(local_path, 'json')) or overwrite:
            # create local path to download and save to
            if not os.path.isdir(local_path):
                os.makedirs(local_path)

        try:
            payload = dict()
            if filters is not None:
                payload['itemsQuery'] = filters.prepare()
            payload['annotations'] = {
                "include": include_annotations_in_output,
                "convertSemantic": export_png_files
            }
            payload['exportVersion'] = export_version
            if annotation_filters is not None:
                payload['annotationsQuery'] = annotation_filters.prepare()
                payload['annotations']['filter'] = filter_output_annotations
            if dataset_lock:
                payload['datasetLock'] = dataset_lock

            if export_summary:
                payload['summary'] = export_summary

            if lock_timeout_sec:
                payload['lockTimeoutSec'] = lock_timeout_sec

            success, response = dataset._client_api.gen_request(req_type='post',
                                                                path='/datasets/{}/export'.format(dataset.id),
                                                                json_req=payload,
                                                                headers={'user_query': filters._user_query})
            if not success:
                raise exceptions.PlatformException(response)
            command = entities.Command.from_json(_json=response.json(),
                                                 client_api=dataset._client_api)
            command = command.wait(timeout=0)
            if 'outputItemId' not in command.spec:
                raise exceptions.PlatformException(
                    error='400',
                    message="outputItemId key is missing in command response: {}".format(response))
            item_id = command.spec['outputItemId']
            annotation_zip_item = repositories.Items(client_api=dataset._client_api).get(item_id=item_id)
            zip_filepath = annotation_zip_item.download(local_path=local_path, export_version=export_version)
            # unzipping annotations to directory
            if isinstance(zip_filepath, list) or not os.path.isfile(zip_filepath):
                raise exceptions.PlatformException(
                    error='404',
                    message='error downloading annotation zip file. see above for more information. item id: {!r}'.format(
                        annotation_zip_item.id))
            try:
                miscellaneous.Zipping.unzip_directory(zip_filename=zip_filepath,
                                                      to_directory=local_path)
            except Exception as e:
                logger.warning("Failed to extract zip file error: {}".format(e))

        finally:
            # cleanup
            if isinstance(zip_filepath, str) and os.path.isfile(zip_filepath):
                os.remove(zip_filepath)

    @staticmethod
    def _download_img_annotations(item: entities.Item,
                                  img_filepath,
                                  local_path,
                                  overwrite,
                                  annotation_options,
                                  annotation_filters,
                                  thickness=1,
                                  with_text=False,
                                  alpha=1,
                                  export_version=entities.ExportVersion.V1
                                  ):

        # check if local_path is a file name
        _, ext = os.path.splitext(local_path)
        if ext:
            # take the dir of the file for the annotations save
            local_path = os.path.dirname(local_path)

        # fix local path
        if local_path.endswith("/items") or local_path.endswith("\\items"):
            local_path = os.path.dirname(local_path)

        annotation_rel_path = item.filename[1:]
        if img_filepath is not None:
            dir_name = os.path.dirname(annotation_rel_path)
            base_name = os.path.basename(img_filepath)
            annotation_rel_path = os.path.join(dir_name, base_name)

        # find annotations json
        annotations_json_filepath = os.path.join(local_path, "json", annotation_rel_path)
        if export_version == entities.ExportVersion.V1:
            name, _ = os.path.splitext(annotations_json_filepath)
        else:
            name = annotations_json_filepath
        annotations_json_filepath = name + ".json"

        if os.path.isfile(annotations_json_filepath) and annotation_filters is None:
            # if exists take from json file
            with open(annotations_json_filepath, "r", encoding="utf8") as f:
                data = json.load(f)
            if "annotations" in data:
                data = data["annotations"]
            annotations = entities.AnnotationCollection.from_json(_json=data, item=item)
            # no need to use the filters here because the annotations were already downloaded with annotation_filters
        else:
            # if json file doesnt exist get the annotations from platform
            annotations = item.annotations.list(filters=annotation_filters)

        # get image shape
        is_url_item = item.metadata. \
                          get('system', dict()). \
                          get('shebang', dict()). \
                          get('linkInfo', dict()). \
                          get('type', None) == 'url'

        if item is not None:
            orientation = item.system.get('exif', {}).get('Orientation', 0)
        else:
            orientation = 0
        if item.width is not None and item.height is not None:
            if orientation in [5, 6, 7, 8]:
                img_shape = (item.width, item.height)
            else:
                img_shape = (item.height, item.width)
        elif ('image' in item.mimetype and img_filepath is not None) or \
                (is_url_item and img_filepath is not None):
            img_shape = Image.open(img_filepath).size[::-1]
        else:
            img_shape = (0, 0)

        # download all annotation options
        for option in annotation_options:
            # get path and create dirs
            annotation_filepath = os.path.join(local_path, option, annotation_rel_path)
            if not os.path.isdir(os.path.dirname(annotation_filepath)):
                os.makedirs(os.path.dirname(annotation_filepath), exist_ok=True)

            if export_version == entities.ExportVersion.V1:
                temp_path, ext = os.path.splitext(annotation_filepath)
            else:
                temp_path = annotation_filepath

            if option == entities.ViewAnnotationOptions.JSON:
                if not os.path.isfile(annotations_json_filepath):
                    annotations.download(
                        filepath=annotations_json_filepath,
                        annotation_format=option,
                        height=img_shape[0],
                        width=img_shape[1],
                    )
            elif option in [entities.ViewAnnotationOptions.MASK,
                            entities.ViewAnnotationOptions.INSTANCE,
                            entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE,
                            entities.ViewAnnotationOptions.OBJECT_ID,
                            entities.ViewAnnotationOptions.VTT]:
                if option == entities.ViewAnnotationOptions.VTT:
                    annotation_filepath = temp_path + ".vtt"
                else:
                    if 'video' in item.mimetype:
                        annotation_filepath = temp_path + ".mp4"
                    else:
                        annotation_filepath = temp_path + ".png"
                if not os.path.isfile(annotation_filepath) or overwrite:
                    # if not exists OR (exists AND overwrite)
                    if not os.path.exists(os.path.dirname(annotation_filepath)):
                        # create folder if not exists
                        os.makedirs(os.path.dirname(annotation_filepath), exist_ok=True)
                    if option == entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE and img_filepath is None:
                        raise PlatformException(
                            error="1002",
                            message="Missing image for annotation option dl.ViewAnnotationOptions.ANNOTATION_ON_IMAGE")
                    annotations.download(
                        filepath=annotation_filepath,
                        img_filepath=img_filepath,
                        annotation_format=option,
                        height=img_shape[0],
                        width=img_shape[1],
                        thickness=thickness,
                        alpha=alpha,
                        with_text=with_text,
                        orientation=orientation
                    )
            else:
                raise PlatformException(error="400", message="Unknown annotation option: {}".format(option))

    @staticmethod
    def __get_local_filepath(local_path, item, to_items_folder, without_relative_path=None, is_folder=False):
        # create paths
        _, ext = os.path.splitext(local_path)
        if ext and not is_folder:
            # local_path is a filename
            local_filepath = local_path
            local_path = os.path.dirname(local_filepath)
        else:
            # if directory - get item's filename
            if to_items_folder:
                local_path = os.path.join(local_path, "items")
            elif is_folder:
                local_path = os.path.join(local_path, "")
            if without_relative_path is not None:
                local_filepath = os.path.join(local_path, item.name)
            else:
                local_filepath = os.path.join(local_path, item.filename[1:])
        return local_path, local_filepath

    @staticmethod
    def __get_link_source(item):
        assert isinstance(item, entities.Item)
        is_url = False
        url = item.resolved_stream
        if item.metadata.get('system', {}).get('shebang', {}).get('linkInfo', {}).get('type', '') == 'url':
            is_url = True

        return item, url, is_url, url.startswith('file://')

    def __file_validation(self, item, downloaded_file):
        res = False
        resume = True
        if isinstance(downloaded_file, io.BytesIO):
            file_size = downloaded_file.getbuffer().nbytes
        else:
            file_size = os.stat(downloaded_file).st_size
        expected_size = item.metadata['system']['size']
        size_diff = file_size - expected_size
        if size_diff == 0:
            res = True
        if size_diff > 0:
            resume = False
        return res, file_size, resume

    def __thread_download(self,
                          item,
                          save_locally,
                          local_path,
                          to_array,
                          local_filepath,
                          overwrite,
                          annotation_options,
                          annotation_filters,
                          chunk_size=8192,
                          thickness=1,
                          with_text=False,
                          alpha=1,
                          export_version=entities.ExportVersion.V1
                          ):
        """
        Get a single item's binary data
        Calling this method will returns the item body itself , an image for example with the proper mimetype.
  
        :param item: Item entity to download
        :param save_locally: bool. save to file or return buffer
        :param local_path: item local folder to save to.
        :param to_array: returns Ndarray when True and local_path = False
        :param local_filepath: item local filepath
        :param overwrite: overwrite the file is existing
        :param annotation_options: download annotations options: list(dl.ViewAnnotationOptions)
        :param annotation_filters: Filters entity to filter item's annotation
        :param chunk_size: size of chunks to download - optional. default = 8192
        :param thickness: optional - line thickness, if -1 annotation will be filled, default =1
        :param with_text: optional - add text to annotations, default = False
        :param alpha: opacity value [0 1], default 1
        :param ExportVersion export_version:  exported items will have original extension in filename, `V1` - no original extension in filenames
        :return:
        """
        # check if need to download image binary from platform
        need_to_download = True
        if save_locally and os.path.isfile(local_filepath):
            need_to_download = overwrite

        item, url, is_url, is_local_link = self.__get_link_source(item=item)

        # save as byte stream
        data = io.BytesIO()
        if need_to_download:
            chunk_resume = {0: 0}
            start_point = 0
            download_done = False
            while chunk_resume.get(start_point, '') != 3 and not download_done:
                if not is_url:
                    headers = {'x-dl-sanitize': '0', 'Range': 'bytes={}-'.format(start_point)}
                    result, response = self.items_repository._client_api.gen_request(req_type="get",
                                                                                     headers=headers,
                                                                                     path="/items/{}/stream".format(
                                                                                         item.id),
                                                                                     stream=True,
                                                                                     dataset_id=item.dataset_id)
                    if not result:
                        if os.path.isfile(local_filepath + '.download'):
                            os.remove(local_filepath + '.download')
                        raise PlatformException(response)
                else:
                    _, ext = os.path.splitext(item.metadata['system']['shebang']['linkInfo']['ref'].split('?')[0])
                    if local_filepath:
                        local_filepath += ext
                    response = self.get_url_stream(url=url)

                if save_locally:
                    # save to file
                    if not os.path.exists(os.path.dirname(local_filepath)):
                        # create folder if not exists
                        os.makedirs(os.path.dirname(local_filepath), exist_ok=True)

                    # decide if create progress bar for item
                    if not is_local_link:
                        total_length = response.headers.get("content-length")
                    else:
                        response.seek(0, 2)
                        total_length = response.tell()
                        response.seek(0)
                    one_file_pbar = None
                    try:
                        one_file_progress_bar = total_length is not None and int(
                            total_length) > 10e6  # size larger than 10 MB
                        if one_file_progress_bar:
                            one_file_pbar = tqdm.tqdm(total=int(total_length),
                                                      unit='B',
                                                      unit_scale=True,
                                                      unit_divisor=1024,
                                                      position=1,
                                                      file=sys.stdout,
                                                      disable=self.items_repository._client_api.verbose.disable_progress_bar_download_item,
                                                      desc='Download Item')
                    except Exception as err:
                        one_file_progress_bar = False
                        logger.debug('Cant decide downloaded file length, bar will not be presented: {}'.format(err))

                    # start download
                    if self.items_repository._client_api.sdk_cache.use_cache and \
                            self.items_repository._client_api.cache is not None:
                        response_output = os.path.normpath(response.content)
                        if isinstance(response_output, bytes):
                            response_output = response_output.decode('utf-8')[1:-1]

                        if os.path.isfile(os.path.normpath(response_output)):
                            if response_output != local_filepath:
                                source_path = os.path.normpath(response_output)
                                shutil.copyfile(source_path, local_filepath)
                        download_done = True
                    else:
                        try:
                            temp_file_path = local_filepath + '.download'
                            with open(temp_file_path, "ab") as f:
                                try:
                                    if is_local_link and isinstance(response, io.BufferedReader):
                                        generator = iter(lambda: response.read(chunk_size), b'')
                                    else:
                                        generator = response.iter_content(chunk_size=chunk_size)
                                    for chunk in generator:
                                        if chunk:  # filter out keep-alive new chunks
                                            f.write(chunk)
                                            if one_file_progress_bar:
                                                one_file_pbar.update(len(chunk))
                                except Exception as err:
                                    pass
                                finally:
                                    if is_local_link and isinstance(response, io.BufferedReader):
                                        try:
                                            response.close()
                                        except Exception as err:
                                            pass

                            file_validation = True
                            if not is_url:
                                file_validation, start_point, chunk_resume = self.__get_next_chunk(
                                    item=item,
                                    download_progress=temp_file_path,
                                    chunk_resume=chunk_resume
                                )
                            if file_validation:
                                shutil.move(temp_file_path, local_filepath)
                                download_done = True
                        except Exception as err:
                            if os.path.isfile(temp_file_path):
                                os.remove(temp_file_path)
                            raise err
                    if one_file_progress_bar:
                        one_file_pbar.close()
                    # save to output variable
                    data = local_filepath
                    # if image - can download annotation mask
                    if item.annotated and annotation_options:
                        self._download_img_annotations(item=item,
                                                       img_filepath=local_filepath,
                                                       annotation_options=annotation_options,
                                                       annotation_filters=annotation_filters,
                                                       local_path=local_path,
                                                       overwrite=overwrite,
                                                       thickness=thickness,
                                                       alpha=alpha,
                                                       with_text=with_text,
                                                       export_version=export_version
                                                       )
                else:
                    if self.items_repository._client_api.sdk_cache.use_cache and \
                            self.items_repository._client_api.cache is not None:
                        response_output = os.path.normpath(response.content)
                        if isinstance(response_output, bytes):
                            response_output = response_output.decode('utf-8')[1:-1]

                        if os.path.isfile(response_output):
                            source_file = response_output
                            with open(source_file, 'wb') as f:
                                data = f.read()
                        download_done = True
                    else:
                        try:
                            if is_local_link and isinstance(response, io.BufferedReader):
                                generator = iter(lambda: response.read(chunk_size), b'')
                            else:
                                generator = response.iter_content(chunk_size=chunk_size)
                            for chunk in generator:
                                if chunk:  # filter out keep-alive new chunks
                                    data.write(chunk)

                            file_validation = True
                            if not is_url:
                                file_validation, start_point, chunk_resume = self.__get_next_chunk(item=item,
                                                                                                   download_progress=data,
                                                                                                   chunk_resume=chunk_resume)
                            if file_validation:
                                download_done = True
                            else:
                                continue
                        except Exception as err:
                            raise err
                        finally:
                            if is_local_link and isinstance(response, io.BufferedReader):
                                try:
                                    response.close()
                                except Exception as err:
                                    pass
                    # go back to the beginning of the stream
                    data.seek(0)
                    data.name = item.name
                    if not save_locally and to_array:
                        if 'image' not in item.mimetype and not is_url:
                            raise PlatformException(
                                error="400",
                                message='Download element type numpy.ndarray support for image only. '
                                        'Item Id: {} is {} type'.format(item.id, item.mimetype))

                        data = np.array(Image.open(data))
        else:
            data = local_filepath
        return data

    def __get_next_chunk(self, item, download_progress, chunk_resume):
        size_validation, file_size, resume = self.__file_validation(item=item,
                                                                    downloaded_file=download_progress)
        start_point = file_size
        if not size_validation:
            if chunk_resume.get(start_point, None) is None:
                chunk_resume = {start_point: 1}
            else:
                chunk_resume[start_point] += 1
            if chunk_resume[start_point] == 3 or not resume:
                raise PlatformException(
                    error=500,
                    message='The downloaded file is corrupted. Please try again. If the issue repeats please contact support.')
        return size_validation, start_point, chunk_resume

    def __default_local_path(self):

        # create default local path
        if self.items_repository._dataset is None:
            local_path = os.path.join(
                self.items_repository._client_api.sdk_cache.cache_path_bin,
                "items",
            )
        else:
            if self.items_repository.dataset._project is None:
                # by dataset name
                local_path = os.path.join(
                    self.items_repository._client_api.sdk_cache.cache_path_bin,
                    "datasets",
                    "{}_{}".format(self.items_repository.dataset.name, self.items_repository.dataset.id),
                )
            else:
                # by dataset and project name
                local_path = os.path.join(
                    self.items_repository._client_api.sdk_cache.cache_path_bin,
                    "projects",
                    self.items_repository.dataset.project.name,
                    "datasets",
                    self.items_repository.dataset.name,
                )
        logger.info("Downloading to: {}".format(local_path))
        return local_path

    @staticmethod
    def get_url_stream(url):
        """
        :param url:
        """
        response = None

        if url.startswith('file://'):
            parsed = urlparse(url)
            path = unquote(parsed.path)
            if parsed.netloc:
                path = f"/{parsed.netloc}{path}"
            path = Path(path).expanduser().resolve()

            if not path.exists():
                raise PlatformException(
                    error='404',
                    message=f'Local file not found: {url}'
                )
            if not path.is_file():
                raise PlatformException(
                    error='400',
                    message=f'Path is not a file: {url}'
                )

            try:
                response = io.BufferedReader(io.FileIO(path, 'rb'))
            except PermissionError as e:
                raise PlatformException(
                    error='403',
                    message=f'Permission denied accessing file: {url}'
                ) from e
        else:
            prepared_request = requests.Request(method='GET', url=url).prepare()
            with requests.Session() as s:
                retry = Retry(
                    total=3,
                    read=3,
                    connect=3,
                    backoff_factor=1,
                )
                adapter = HTTPAdapter(max_retries=retry)
                s.mount('http://', adapter)
                s.mount('https://', adapter)
                response = s.send(request=prepared_request, stream=True)

        return response



import tempfile
import requests
import logging
import shutil
import os

from .. import entities, miscellaneous, PlatformException, exceptions, repositories
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Artifacts:
    """
    Artifacts repository
    """

    def __init__(self,
                 client_api: ApiClient,
                 project: entities.Project = None,
                 dataset: entities.Dataset = None,
                 project_id: str = None,
                 model: entities.Model = None,
                 package: entities.Package = None,
                 dataset_name=None):
        self._client_api = client_api
        self._project = project
        self._dataset = dataset
        self._items_repository = None
        self.dataset_name = dataset_name
        self.project_id = project_id
        self.model = model
        self.package = package

    ############
    # entities #
    ############
    @property
    def dataset(self) -> entities.Dataset:
        if self._dataset is None:
            # get dataset from project
            try:
                if self.dataset_name is None:
                    self.dataset_name = 'Binaries'
                    self._dataset = self.project.datasets._get_binaries_dataset()
                else:
                    self._dataset = self.project.datasets.get(dataset_name=self.dataset_name)
            except exceptions.NotFound:
                raise ValueError(
                    f'Missing "{self.dataset_name}" dataset in the project. Please contact support for help')

        return self._dataset

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            if self._dataset is not None:
                self._project = self._dataset.project
            elif self.project_id is not None:
                self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id)
            else:
                raise PlatformException(error='400',
                                        message='Artifacts doesnt have a project, at least one must not be None: dataset, project or project_id')
        return self._project

    ################
    # repositories #
    ################
    @property
    def items_repository(self):
        if self._items_repository is None:
            # load Binaries dataset
            # load items repository
            self._items_repository = self.dataset.items
            self._items_repository.set_items_entity(entities.Artifact)
        return self._items_repository

    ###########
    # methods #
    ###########
    @staticmethod
    def _build_path_header(
            package_name=None,
            package=None,
            execution_id=None,
            execution=None,
            model_name=None,
    ):
        remote_path = '/artifacts'
        if package_name is not None or package is not None:
            if package is not None:
                package_name = package.name
            remote_path += '/packages/{}'.format(package_name)
        if execution_id is not None or execution is not None:
            if execution is not None:
                execution_id = execution.id
            remote_path += '/executions/{}'.format(execution_id)
        if model_name is not None:
            remote_path += '/models/{}'.format(model_name)

        return remote_path

    def list(self,
             execution_id: str = None,
             package_name: str = None,
             model_name: str = None) -> miscellaneous.List[entities.Artifact]:
        """
        List of artifacts

        :param str execution_id: list by execution id
        :param str package_name: list by package name
        :param str model_name: list by model name
        :return: list of artifacts
        :rtype: miscellaneous.List[dtlpy.entities.artifact.Artifact]

        **Example**:

        .. code-block:: python

            project.artifacts.list(package_name='package_name')
        """
        if self.model is not None:
            model_name = self.model.name
        if self.package is not None:
            package_name = self.package.name

        filters = entities.Filters()
        remote_path = self._build_path_header(
            package_name=package_name,
            execution_id=execution_id,
            model_name=model_name
        )

        remote_path += '/*'
        filters.add(field='filename', values=remote_path)
        pages = self.items_repository.list(filters=filters)
        items = [entities.Artifact.from_json(_json=item.to_json(),
                                             client_api=self._client_api,
                                             dataset=item._dataset,
                                             project=self.project)
                 for page in pages for item in page]
        return miscellaneous.List(items)

    def get(self,
            artifact_id: str = None,
            artifact_name: str = None,
            model_name: str = None,
            execution_id: str = None,
            package_name: str = None) -> entities.ItemArtifact:
        """

        Get an artifact object by name, id or type
        If by name or type - need to input also execution/task id for the artifact folder

        :param str artifact_id: search by artifact id
        :param str artifact_name: search by artifact id
        :param str model_name: search by model name
        :param str execution_id: search by execution id
        :param str package_name: search by  package name
        :return: Artifact object
        :rtype: dtlpy.entities.artifact.Artifact

        **Example**:

        .. code-block:: python

            project.artifacts.get(artifact_id='artifact_id')
        """
        if self.model is not None:
            model_name = self.model.name
        if self.package is not None:
            package_name = self.package.name

        if artifact_id is not None:
            artifact = self.items_repository.get(item_id=artifact_id)
            # verify input artifact name is same as the given id
            if artifact_name is not None and artifact.name != artifact_name:
                logger.warning(
                    "Mismatch found in artifacts.get: artifact_name is different then artifact.name:"
                    " {!r} != {!r}".format(
                        artifact_name,
                        artifact.name))
        elif artifact_name is not None:
            artifacts = self.list(
                execution_id=execution_id,
                package_name=package_name,
                model_name=model_name
            )
            artifact = [artifact for artifact in artifacts if artifact.name == artifact_name]
            if len(artifact) == 1:
                artifact = artifact[0]
            elif len(artifact) > 1:
                raise PlatformException('404', 'More Than one Artifact found')
            else:
                raise PlatformException('404', 'Artifact not found')
        else:
            msg = 'one input must be not None: artifact_id or artifact_name'
            raise ValueError(msg)
        return artifact

    def download(
            self,
            artifact_id: str = None,
            artifact_name: str = None,
            execution_id: str = None,
            package_name: str = None,
            model_name: str = None,
            local_path: str = None,
            overwrite: bool = False,
            save_locally: bool = True
    ):
        """

        Download artifact binary.
        Get artifact by name, id or type

        :param str artifact_id: search by artifact id
        :param str artifact_name: search by artifact id
        :param str execution_id: search by execution id
        :param str package_name: search by package name
        :param str model_name: search by model name
        :param str local_path: artifact will be saved to this filepath
        :param bool overwrite: optional - default = False
        :param bool save_locally: to save the file local
        :return: file path
        :rtype: str

        **Example**:

        .. code-block:: python

            project.artifacts.download(artifact_id='artifact_id',
                                        local_path='your_path',
                                        overwrite=True,
                                        save_locally=True)
        """
        if self.model is not None:
            model_name = self.model.name
        if self.package is not None:
            package_name = self.package.name

        if artifact_id is not None:
            # download
            artifact = self.items_repository.download(items=artifact_id,
                                                      save_locally=save_locally,
                                                      local_path=local_path,
                                                      overwrite=overwrite)
        elif artifact_name is not None:
            artifact_obj: entities.ItemArtifact = self.get(artifact_id=artifact_id,
                                                           execution_id=execution_id,
                                                           package_name=package_name,
                                                           artifact_name=artifact_name)

            artifact = artifact_obj.download(save_locally=save_locally,
                                             local_path=local_path,
                                             overwrite=overwrite)

        else:
            if self.model is not None:
                artifact = list()
                for m_artifact in self.model.model_artifacts:
                    if isinstance(m_artifact, entities.ItemArtifact):
                        if not m_artifact.is_fetched:
                            m_artifact = self.items_repository.get(item_id=m_artifact.id)
                        model_remote_root = m_artifact.filename.split('/')
                        model_remote_root = '/'.join(model_remote_root[:4])
                        # remove the prefix with relpath
                        local_dst = os.path.join(local_path,
                                                 os.path.relpath(m_artifact.filename, model_remote_root))
                        if not os.path.isfile(local_dst) or overwrite:
                            # need_to_download
                            # 1. download to temp folder
                            temp_dir = tempfile.mkdtemp()
                            local_temp_file = m_artifact.download(
                                local_path=temp_dir,
                                overwrite=overwrite,
                                to_items_folder=False,
                            )
                            src = local_temp_file
                            # remove the prefix with relpath
                            dst = local_dst
                            os.makedirs(os.path.dirname(dst), exist_ok=True)
                            shutil.move(src=src, dst=dst)
                            # clean temp dir
                            if os.path.isdir(temp_dir):
                                shutil.rmtree(temp_dir)
                        artifact.append(local_path)
                    elif isinstance(m_artifact, entities.LinkArtifact):
                        # remove the prefix with relpath
                        local_dst = os.path.join(local_path, m_artifact.filename)
                        if not os.path.isfile(local_dst) or overwrite:
                            # need_to_download
                            # 1. download to temp folder
                            temp_dir = tempfile.mkdtemp()
                            response = requests.get(m_artifact.url, stream=True)
                            local_temp_file = os.path.join(temp_dir, m_artifact.filename)
                            with open(local_temp_file, "wb") as handle:
                                for data in response.iter_content(chunk_size=8192):
                                    handle.write(data)
                            src = local_temp_file
                            # remove the prefix with relpath
                            dst = local_dst
                            os.makedirs(os.path.dirname(dst), exist_ok=True)
                            shutil.move(src=src, dst=dst)
                            # clean temp dir
                            if os.path.isdir(temp_dir):
                                shutil.rmtree(temp_dir)
                        artifact.append(local_path)
                    elif isinstance(m_artifact, entities.LocalArtifact):
                        pass
                    else:
                        raise ValueError('unsupported artifact type: {}'.format(type(m_artifact)))
            else:
                # for package artifacts - download using filter on the package directory
                if all(elem is None for elem in [package_name, execution_id]):
                    raise PlatformException(error='400', message='Must input package or execution (id or entity)')
                remote_path = self._build_path_header(
                    package_name=package_name,
                    execution_id=execution_id,
                    model_name=model_name,
                )
                without_relative_path = remote_path
                remote_path += '/*'
                filters = entities.Filters()
                filters.add(field='filename', values=remote_path)
                artifact = self.items_repository.download(filters=filters,
                                                          save_locally=save_locally,
                                                          local_path=local_path,
                                                          to_items_folder=False,
                                                          overwrite=overwrite,
                                                          without_relative_path=without_relative_path)
        return artifact

    def upload(self,
               # what to upload
               filepath: str,
               # where to upload
               package_name: str = None,
               package: entities.Package = None,
               execution_id: str = None,
               execution: entities.Execution = None,
               model_name: str = None,
               # add information
               overwrite: bool = False):
        """

        Upload binary file to artifact. get by name, id or type.
        If artifact exists - overwriting binary
        Else and if create==True a new artifact will be created and uploaded

        :param str filepath: local binary file
        :param str package_name: package name that include the artifact
        :param dtlpy.entities.package.Package package: package object
        :param str execution_id: execution id that include the artifact
        :param dtlpy.entities.execution.Execution execution: execution object
        :param str model_name: model name that include the artifact
        :param bool overwrite: optional - default = False to overwrite an existing object
        :return: Artifact Object
        :rtype: dtlpy.entities.artifact.Artifact

        **Example**:

        .. code-block:: python

            project.artifacts.upload(filepath='filepath',
                                    package_name='package_name')
        """
        if self.model is not None:
            model_name = self.model.name
        if self.package is not None:
            package_name = self.package.name

        remote_path = self._build_path_header(package_name=package_name,
                                              package=package,
                                              execution=execution,
                                              execution_id=execution_id,
                                              model_name=model_name)

        if all(elem is None for elem in [package_name, package, execution_id, execution, model_name]):
            raise ValueError('Must input package or execution (id or entity)')

        artifact = self.items_repository.upload(local_path=filepath,
                                                remote_path=remote_path,
                                                overwrite=overwrite,
                                                output_entity=entities.Artifact)
        if self.model is not None:
            # list and update model
            filters = entities.Filters()
            filters.add(field='dir', values=remote_path + '*')
            pages = self.items_repository.list(filters=filters)
            model_artifacts = list()
            for item in pages.all():
                model_artifacts.append(entities.Artifact.from_json(_json=item.to_json(),
                                                                   client_api=self._client_api,
                                                                   dataset=item._dataset))
            self.model.model_artifacts = model_artifacts
            self.model.update()
        logger.debug('Artifact uploaded successfully')
        return artifact

    def delete(self,
               artifact_id=None,
               artifact_name=None,
               execution_id=None,
               model_name=None,
               package_name=None):
        """
        Delete artifacts

        :param str artifact_id: search by artifact id
        :param str artifact_name: search by artifact id
        :param str execution_id: search by execution id
        :param str model_name: search by model name
        :param str package_name: search by package name
        :return: True if success
        :rtype: bool

         **Example**:

        .. code-block:: python

            project.artifacts.delete(artifact_id='artifact_id',
                                    package_name='package_name')
        """
        if self.model is not None:
            model_name = self.model.name
        if self.package is not None:
            package_name = self.package.name

        if artifact_id is not None or artifact_name is not None:
            artifacts = [
                self.get(
                    artifact_id=artifact_id,
                    artifact_name=artifact_name,
                    model_name=model_name,
                )
            ]
        elif execution_id is not None or package_name is not None:
            artifacts = self.list(
                execution_id=execution_id,
                package_name=package_name,
                model_name=model_name,
            )
        else:
            raise PlatformException('400',
                                    'Must provide one of: artifact_id, artifact_name, execution_id, package_name')

        values = [artifact.id for artifact in artifacts]
        self.items_repository.delete(filters=entities.Filters(field='id', values=values,
                                                              operator=entities.FiltersOperations.IN))

        return True



from .. import entities
from ..services.api_client import ApiClient
import attr

@attr.s
class Collection(entities.BaseEntity):
    """
    Represents a collection in the dataset.
    """

    # sdk
    _client_api = attr.ib(type=ApiClient, repr=False)

    key = attr.ib(type=str)
    name = attr.ib(type=str)

    @classmethod
    def from_json(cls, _json, client_api, is_fetched=True):
        """
        Create a single Collection entity from the dataset JSON.

        :param _json: A dictionary containing collection data in the format:
                    { "metadata.system.collections.c0": {"name": "Justice League"} }
        :param client_api: The client API instance.
        :param is_fetched: Whether the entity was fetched from the platform.
        :return: A single Collection entity.
        """
        full_key, value = next(iter(_json.items()))
        # Strip the prefix
        key = full_key.replace("metadata.system.collections.", "")
        name = value.get("name")

        inst = cls(
            key=key,
            name=name,
            client_api=client_api,
        )
        inst.is_fetched = is_fetched
        return inst


import json
import os
import shutil
import time
from enum import Enum
from pathlib import Path
import mmap
from filelock import FileLock
import logging
import base64

from .dl_cache import DiskCache
from .redis_cache import RedisCache
from .filesystem_cache import FileSystemCache

logger = logging.getLogger(name='dtlpy')


class ObjectType(str, Enum):
    BINARY = "binary"
    OBJECT = "object"


class CacheType(Enum):
    DISKCACHE = 'diskcache'
    REDIS = 'redis'
    FILESYSTEM = 'filesystem'


class CacheConfig:
    def __init__(self, cache_type=CacheType.DISKCACHE, ttl=1000, level=1, options=None):
        """
        Cache config settings

        :param CacheType cache_type: CacheType diskcache, filesystem, redis
        :param int ttl: time to hold the item in the cache in seconds (SEC)
        :param int level: cache level
        :param dict options: the configs for the caches types
        """
        if isinstance(cache_type, CacheType):
            cache_type = cache_type.value
        if isinstance(cache_type, str) and cache_type not in CacheType._value2member_map_:
            raise ValueError('cache type must be redis or diskcache')

        self.type = cache_type
        self.ttl = ttl
        self.level = level
        self.options = options

    def to_string(self):
        """
        convert object to base 64 string
        """
        base64_bytes = base64.b64encode(json.dumps(self.to_json()).encode("ascii"))
        base64_string = base64_bytes.decode("ascii")
        return base64_string

    @staticmethod
    def from_string(cls, base64_string):
        """
        convert from base 64 string to the class object

        :param str base64_string: string in base64 the have a json configs
        """
        base64_bytes = base64_string.encode("ascii")
        sample_string_bytes = base64.b64decode(base64_bytes)
        _json = json.loads(sample_string_bytes.decode("ascii"))
        return cls(cache_type=_json.get('type', CacheType.DISKCACHE),
                   ttl=_json.get('ttl', 1000),
                   level=_json.get('level', 1),
                   options=_json.get('options', None))

    def to_json(self):
        """
        convert the class to json
        """
        return {
            'type': self.type,
            'ttl': self.ttl,
            'level': self.level,
            'options': self.options,
        }

    @staticmethod
    def from_json(cls, _json):
        """
        make a class attribute from json

        :param _json: _json have the class attributes
        """
        if isinstance(_json, str):
            _json = json.loads(_json)
        return cls(cache_type=_json.get('type', CacheType.DISKCACHE),
                   ttl=_json.get('ttl', 1000),
                   level=_json.get('level', 1),
                   options=_json.get('options', None))


class CacheKey:
    def __init__(self,
                 master_type='**',
                 master_id='**',
                 entity_type='**',
                 entity_id='*',
                 object_type=ObjectType.OBJECT):
        """
        :param str master_type: master type
        :param str master_id: master id
        :param str entity_type: entity type
        :param str entity_id: entity id
        :param str object_type: object type object/binary
        """
        self.master_type = master_type
        self.master_id = master_id
        self.entity_type = entity_type
        self.entity_id = entity_id
        self.object_type = object_type

    def get(self):
        """
        return the build key
        """
        return os.path.join(self.master_type, self.master_id, self.entity_type, self.entity_id, self.object_type)

    def get_key(self):
        """
        return the build key
        """
        return os.path.join(self.entity_type, self.entity_id, self.object_type)


class CacheManger:
    def __init__(self, cache_configs: list, bin_cache_size=1000):
        """
        Cache manger for config and mange the cache

        :param cache_configs: CacheConfig object
        :param bin_cache_size: size on MB for binary cache
        """
        self.cache_levels = dict()
        self._max_level = 1
        self.bin_cache_size = bin_cache_size
        self.bin_cache_path = os.environ['DEFAULT_CACHE_PATH']
        self._current_bin_cache_size = 0
        for config in cache_configs:
            try:
                self.cache_levels[config.level] = self._load_cache_handler(config)
                if config.level < self._max_level:
                    self._max_level = config.level
            except:
                raise "Failed to build Cache"

        self.parent_dict = {
            "annotations": 'items',
            "items": 'datasets',
            "datasets": 'projects',
            "projects": 'org',
            "org": '',
            "annotationtasks": 'datasets',
            "assignments": 'annotationtasks',
            "models": 'packages',
            "packages": 'projects',
            "services": 'packages',
        }

    def _load_cache_handler(self, config: CacheConfig):
        """
        the function the build the cache form the configs that get
        """
        from ..services import DataloopLogger
        cache = None
        if config.type == CacheType.REDIS.value:
            try:
                cache = RedisCache(options=config.options, ttl=config.ttl)
            except:
                logger.warning("Failed to build Redis")
                raise Exception("Failed to build Redis")

        elif config.type == CacheType.DISKCACHE.value:
            cache = DiskCache(name='object_cache', options=config.options, ttl=config.ttl)
        elif config.type == CacheType.FILESYSTEM.value:
            cache = FileSystemCache(options=config.options, ttl=config.ttl)
            DataloopLogger.clean_dataloop_cache(cache_path=cache.root_dir,
                                                max_param={'max_time': cache.ttl})
        DataloopLogger.clean_dataloop_cache(cache_path=self.bin_cache_path,
                                            max_param={'max_time': config.ttl})
        return cache

    def get(self, key: CacheKey):
        """
        Cache get

        :param CacheKey key: CacheKey object
        :return: success, list of the get result
        """
        res = []
        success = False
        for i in range(1, self._max_level + 1):
            res = self.cache_levels[i].get(key=key.get_key())
            if res:
                success = True
                break
        return success, res

    def ping(self):
        """
        Cache ping check if connection is working
        """
        try:
            for i in range(1, self._max_level + 1):
                self.cache_levels[i].ping()
        except Exception as e:
            raise Exception('cache connection failed ')


    def set(self, key: str, value):
        """
        Cache set, add or update the key value

        :param CacheKey key: CacheKey object
        :param value: value to set
        """
        if isinstance(value, dict):
            value = json.dumps(value)
        self.cache_levels[1].set(key, value)

    def _delete_parent(self, key: CacheKey, level):
        parent_key = CacheKey(master_type=self.parent_dict[key.entity_type],
                              entity_type=key.entity_type,
                              entity_id=key.entity_id,
                              object_type='*')
        list_keys = self.cache_levels[level].list(pattern=parent_key.get())
        for k in list_keys:
            if 'binary' in k:
                val = self.cache_levels[level].get(key=k)
                if os.path.isfile(val):
                    os.remove(val)
            self.cache_levels[level].delete(k)

    def delete(self, key: CacheKey):
        """
        Cache delete

        :param CacheKey key: CacheKey object
        """
        for i in range(1, self._max_level + 1):
            self.cache_levels[i].delete(key.get_key())
            self._delete_parent(key=key, level=i)
            key.object_type = '*'
            list_keys = self.cache_levels[i].list(pattern=key.get_key())
            for k in list_keys:
                val = self.cache_levels[i].get(key=k)
                self.cache_levels[i].delete(k)
                if 'binary' in k:
                    if os.path.isfile(val):
                        os.remove(val)
                    continue
                e_type, e_id, e_obj = val.split('\\')
                self.delete(key=CacheKey(entity_type=e_type, entity_id=e_id, object_type=e_obj))

    def build_cache_key(self, entity_json: dict):
        """
        Build a format of the cache key from the entity json we get

        :param dict entity_json: json of an entity
        :return: CacheKey object
        """
        child_entity = False
        if 'url' in entity_json:
            split_url = entity_json['url'].split('/')
            entity_type = split_url[-2]
            child_entity = True
        elif 'org' in entity_json:
            entity_type = 'projects'
        else:
            entity_type = 'org'
        entity_id = entity_json['id']
        master_type = self.parent_dict[entity_type]
        master_id = '**'
        if child_entity:
            master_id_key = master_type[:-1] + 'Id'
            if master_id_key in entity_json:
                master_id = entity_json[master_id_key]
            elif master_type in entity_json:
                master_id = entity_json[master_type][0]
        elif entity_type == 'projects':
            master_id = entity_json[master_type]['id']

        return CacheKey(master_type=master_type, master_id=master_id, entity_type=entity_type, entity_id=entity_id)

    def _update_config_file(self, filepath: str, update: bool, size: float = 0):
        """
        Update the config file the have all the details about binary cache

        :param str filepath: path of the file the work on
        :param bool update: if True update the use of the file
        :param int size: file size
        """
        config_file_path = os.path.join(self.bin_cache_path, 'cacheConfig.json')
        if os.path.isfile(config_file_path):
            with FileLock(config_file_path + ".lock"):
                with open(config_file_path, mode="r", encoding="utf-8") as con:
                    with mmap.mmap(con.fileno(), length=0, access=mmap.ACCESS_READ) as mmap_obj:
                        text = mmap_obj.read().decode('utf8').replace("'", '"')
                        config_file = json.loads(text)
        else:
            config_file = {'size': 0, 'keys': []}

        if update and filepath in config_file['keys']:
            config_file['keys'].remove(filepath)

        if filepath not in config_file['keys']:
            config_file['keys'].append(filepath)
        config_file['size'] += size
        self._current_bin_cache_size = config_file['size']
        json_object = json.dumps(config_file, indent=4)
        with FileLock(config_file_path + ".lock"):
            with open(config_file_path, mode="w", encoding="utf-8") as outfile:
                outfile.write(json_object)

    def _lru_cache(self):
        """
        Make lru on the binary cache remove 30% of the files
        """
        config_file_path = os.path.join(self.bin_cache_path, 'cacheConfig.json')
        with FileLock(config_file_path + ".lock"):
            with open(config_file_path, mode="r", encoding="utf-8") as con:
                with mmap.mmap(con.fileno(), length=0, access=mmap.ACCESS_READ) as mmap_obj:
                    text = mmap_obj.read().decode('utf8').replace("'", '"')
                    config_file = json.loads(text)

        size = config_file['size']
        end = 70 / 100 * self.bin_cache_size

        while size > end and len(config_file['keys']) > 1:
            to_delete = config_file['keys'][0]

            size -= (Path(to_delete).stat().st_size / 1000000)
            os.remove(to_delete)
            config_file['keys'].remove(to_delete)

        config_file['size'] = size
        json_object = json.dumps(config_file, indent=4)

        with FileLock(config_file_path + ".lock"):
            with open(config_file_path, "w") as outfile:
                outfile.write(json_object)

    def read_stream(self, request_path, dataset_id=None):
        """
        Cache binary get

        :param str request_path: the request
        :param str dataset_id: dataset id of the binary object
        :return: success, list of the get result
        """
        entity_id = request_path.split('/')[-2]
        key = CacheKey(master_type='datasets',
                       master_id=dataset_id,
                       entity_id=entity_id,
                       entity_type='items',
                       object_type=ObjectType.BINARY.value)
        hit, response = self.get(key=key)
        if hit:
            source_path = os.path.normpath(response[0])
            self._update_config_file(filepath=source_path, update=True)
            return hit, [source_path]
        else:
            return False, None

    def write_stream(self,
                     request_path,
                     response=None,
                     buffer=None,
                     file_name=None,
                     entity_id=None,
                     dataset_id=None
                     ):
        """
        Cache binary set

        :param  request_path: the request
        :param  response: the response of stream
        :param  buffer: the steam buffer
        :param  file_name: the file name
        :param  entity_id: entity id
        :param  dataset_id: dataset id of the binary object
        :return: the file path of the binary
        """
        if entity_id is None:
            entity_id = request_path.split('/')[-2]
        key = CacheKey(master_type='datasets',
                       master_id=dataset_id,
                       entity_id=entity_id,
                       entity_type='items',
                       object_type=ObjectType.BINARY)
        filepath = self.bin_cache_path
        if file_name is None:
            file_name = (dict(response.headers)['Content-Disposition'].split('=')[1][2:-1])
        filepath = os.path.join(
            filepath,
            'items',
            file_name
        )
        self.set(key=key.get(), value=filepath)
        if not os.path.isfile(filepath):
            os.makedirs(os.path.dirname(filepath), exist_ok=True)
            if buffer is None:
                try:
                    temp_file_path = filepath + '.download'
                    with open(temp_file_path, "wb") as f:
                        for chunk in response.iter_content(chunk_size=8192):
                            if chunk:  # filter out keep-alive new chunks
                                f.write(chunk)
                    shutil.move(temp_file_path, filepath)
                except:
                    if os.path.isfile(temp_file_path):
                        os.remove(temp_file_path)
                    return ''
            else:
                if os.path.isfile(buffer.name):
                    shutil.copyfile(buffer.name, filepath)
                else:
                    with open(filepath, "wb") as f:
                        f.write(buffer.getbuffer())
        self._update_config_file(filepath=filepath, update=False, size=(Path(filepath).stat().st_size / 1000000))
        if (Path(filepath).stat().st_size / 1000000) + self._current_bin_cache_size > self.bin_cache_size:
            self._lru_cache()
        return filepath

    def read(self, request_path: str):
        """
        Cache entity get

        :param str request_path: the request
        :return: success, list of the get result
        """
        entity_id = request_path.split('/')[-1]
        entity_type = request_path.split('/')[-2]
        key = CacheKey(entity_id=entity_id, entity_type=entity_type)
        hit, response = self.get(key=key)
        if hit:
            return hit, response
        return False, None

    def write(self, list_entities_json):
        """
        Add or update the entity cache

        :param list list_entities_json: list of jsons of entities to set
        """
        for entity_json in list_entities_json:
            key = self.build_cache_key(entity_json)
            redis_key = key.get_key()
            self.set(key=redis_key, value=entity_json)
            self.set(key=key.get(), value=redis_key)

    def invalidate(self, path):
        """
        Delete from the caches

        :param str path: the request path
        """
        entity_id = path.split('/')[-1]
        entity_type = path.split('/')[-2]
        key = CacheKey(entity_id=entity_id, entity_type=entity_type)
        self.delete(key)

    def clear(self):
        self.cache_levels[1].clear()

    def keys(self):
        return [k for k in self.cache_levels[1].keys()]



import os


class PATHS:
    MULTI_METHOD = 'multi_method.py'
    MULTI_METHOD_JSON = 'multi_method_json.py'
    MULTI_METHOD_ITEM = 'multi_method_item.py'
    SINGLE_METHOD_DATASET = 'single_method_dataset.py'
    SINGLE_METHOD = 'single_method.py'
    SINGLE_METHOD_MULTI_INPUT = 'single_method_multi_input.py'
    MULTI_METHOD_DATASET = 'multi_method_dataset.py'
    SINGLE_METHOD_JSON = 'single_method_json.py'
    SINGLE_METHOD_ITEM = 'single_method_item.py'
    SINGLE_METHOD_ANNOTATION = 'single_method_annotation.py'
    MULTI_METHOD_ANNOTATION = 'multi_method_annotation.py'
    CONVERTER = 'converter.py'

    ASSETS_PATH = os.path.dirname(__file__)
    MULTI_METHOD_SR_PATH = os.path.join(ASSETS_PATH, MULTI_METHOD)
    MULTI_METHOD_JSON_SR_PATH = os.path.join(ASSETS_PATH, MULTI_METHOD_JSON)
    MULTI_METHOD_ITEM_SR_PATH = os.path.join(ASSETS_PATH, MULTI_METHOD_ITEM)
    SINGLE_METHOD_DATASET_SR_PATH = os.path.join(ASSETS_PATH, SINGLE_METHOD_DATASET)
    SINGLE_METHOD_SR_PATH = os.path.join(ASSETS_PATH, SINGLE_METHOD)
    MULTI_METHOD_DATASET_SR_PATH = os.path.join(ASSETS_PATH, MULTI_METHOD_DATASET)
    SINGLE_METHOD_JSON_SR_PATH = os.path.join(ASSETS_PATH, SINGLE_METHOD_JSON)
    SINGLE_METHOD_ITEM_SR_PATH = os.path.join(ASSETS_PATH, SINGLE_METHOD_ITEM)
    SINGLE_METHOD_ANNOTATION_SR_PATH = os.path.join(ASSETS_PATH, SINGLE_METHOD_ANNOTATION)
    MULTI_METHOD_ANNOTATION_SR_PATH = os.path.join(ASSETS_PATH, MULTI_METHOD_ANNOTATION)
    SINGLE_METHOD_MULTI_INPUT_SR_PATH = os.path.join(ASSETS_PATH, SINGLE_METHOD_MULTI_INPUT)
    CONVERTER_SR_PATH = os.path.join(ASSETS_PATH, CONVERTER)


service_runner_paths = PATHS()



import datetime
import tabulate
import typing
import logging
import pandas

from .. import exceptions

logger = logging.getLogger(name='dtlpy')

T = typing.TypeVar('T')


class List(list, typing.MutableSequence[T]):
    def to_df(self, show_all=False, columns=None):
        try:
            to_print = list()
            keys_list = list()
            for element in self.__iter__():
                if hasattr(element, 'to_json'):
                    item_dict = element.to_json()
                else:
                    item_dict = element
                to_print.append(item_dict)
                [keys_list.append(key) for key in list(item_dict.keys()) if key not in keys_list]
            try:
                # try sorting bt creation date
                to_print = sorted(to_print, key=lambda k: k['createdAt'] if k['createdAt'] is not None else "")
            except KeyError:
                pass
            except Exception:
                logger.exception('Error sorting printing:')

            remove_keys_list = ['contributors', 'url', 'annotations', 'items', 'export', 'directoryTree', 'org',
                                '_contributors', 'role', 'account', 'featureConstraints',
                                'attributes', 'partitions', 'metadata', 'stream', 'updatedAt', 'arch',
                                'input', 'revisions', 'pipeline',  # task fields
                                'feedbackQueue',  # session fields
                                '_ontology_ids', '_labels',  # dataset
                                'esInstance', 'esIndex',  # time series fields
                                'thumbnail',  # item thumnail too long
                                # services fields
                                'driverId', 'useUserJwt', 'versions', 'runtime', 'mq', 'global',
                                # triggers
                                'scope',
                                # Package
                                'modules'
                                ]
            if not show_all:
                if columns is not None:
                    # take columns from inputs
                    if not isinstance(columns, list):
                        if not isinstance(columns, str):
                            raise exceptions.PlatformException(
                                error='3002',
                                message='"columns" input must be str or list. found: {}'.format(type(columns)))
                        columns = [columns]
                    keys_list = columns
                else:
                    # take default columns
                    for key in remove_keys_list:
                        if key in keys_list:
                            keys_list.remove(key)

            for element in to_print:
                # handle printing errors for not ascii string when in cli
                if 'name' in element:
                    try:
                        # check if ascii
                        element['name'].encode('ascii')
                    except UnicodeEncodeError:
                        # if not - print bytes instead
                        element['name'] = str(element['name']).encode('utf-8')
                if 'createdAt' in element:
                    try:
                        str_timestamp = str(element['createdAt'])
                        if len(str_timestamp) > 10:
                            str_timestamp = str_timestamp[:10]
                        element['createdAt'] = datetime.datetime.fromtimestamp(int(str_timestamp), datetime.timezone.utc).isoformat()
                    except Exception:
                        pass
            df = pandas.DataFrame(to_print, columns=keys_list)
            return df
        except Exception:
            raise exceptions.PlatformException(error='3002',
                                               message='Failed converting to DataFrame')

    def print(self, show_all=False, level='print', to_return=False, columns=None):
        try:
            df = self.to_df(show_all=show_all, columns=columns)
            if 'name' in list(df.columns.values):
                df['name'] = df['name'].astype(str)

            if to_return:
                return tabulate.tabulate(df, headers='keys', tablefmt='psql')
            else:
                if level == 'print':
                    print('\n{}'.format(tabulate.tabulate(df, headers='keys', tablefmt='psql')))
                elif level == 'debug':
                    logger.debug('\n{}'.format(tabulate.tabulate(df, headers='keys', tablefmt='psql')))
                else:
                    raise ValueError('unknown log level in printing: {}'.format(level))

        except Exception:
            raise exceptions.PlatformException(error='3002', message='Failed printing entity')



import dtlpy as dl
import logging

logger = logging.getLogger(name='dtlpy')


class ServiceRunner(dl.BaseServiceRunner):
    """
    Package runner class

    """

    def __init__(self):
        """
        Init package attributes here

        :return:
        """

    def first_method(self, progress=None):
        """
        In this example we print print 'Hello World'

        :param progress: Use this to update the progress of your package
        :return:
        """
        # these lines can be removed
        assert isinstance(progress, dl.Progress)
        progress.update(status='inProgress', progress=0)
        print('Hello World from Dataloop :)')

    def second_method(self, progress=None):
        """
        In this example we print print 'Hello World'

        :param progress: Use this to update the progress of your package
        :return:
        """
        # these lines can be removed
        assert isinstance(progress, dl.Progress)
        progress.update(status='inProgress', progress=0)
        print('Hello World from Dataloop :)')


if __name__ == "__main__":
    """
    Run this main to locally debug your package
    """
    dl.packages.test_local_package()
